@@ -1,31 +1,20 @@
-/********************************************************************
- KWin - the KDE window manager
- This file is part of the KDE project.
+/*
+    KWin - the KDE window manager
+    This file is part of the KDE project.
 
-Copyright (C) 2006-2007 Rivo Laks <rivolaks@hot.ee>
-Copyright (C) 2010, 2011 Martin Gräßlin <mgraesslin@kde.org>
+    SPDX-FileCopyrightText: 2006-2007 Rivo Laks <rivolaks@hot.ee>
+    SPDX-FileCopyrightText: 2010, 2011 Martin Gräßlin <mgraesslin@kde.org>
 
-This program is free software; you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2 of the License, or
-(at your option) any later version.
-
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program.  If not, see <http://www.gnu.org/licenses/>.
-*********************************************************************/
+    SPDX-License-Identifier: GPL-2.0-or-later
+*/
 
 #ifndef KWIN_GLUTILS_H
 #define KWIN_GLUTILS_H
 
 // kwin
-#include <kwinglutils_export.h>
-#include "kwinglutils_funcs.h"
 #include "kwingltexture.h"
+#include "kwinglutils_funcs.h"
+#include <kwinglutils_export.h>
 
 // Qt
 #include <QSize>
@@ -39,8 +28,8 @@ class QVector3D;
 class QVector4D;
 class QMatrix4x4;
 
-template< class K, class V > class QHash;
-
+template<class K, class V>
+class QHash;
 
 namespace KWin
 {
@@ -52,17 +41,16 @@ class GLVertexBufferPrivate;
 //  well as checking for GL version and extensions
 //  Note that GL context has to be created by the time this function is called
 typedef void (*resolveFuncPtr)();
-void KWINGLUTILS_EXPORT initGL(std::function<resolveFuncPtr(const char*)> resolveFunction);
+void KWINGLUTILS_EXPORT initGL(const std::function<resolveFuncPtr(const char *)> &resolveFunction);
 // Cleans up all resources hold by the GL Context
 void KWINGLUTILS_EXPORT cleanupGL();
 
-
 bool KWINGLUTILS_EXPORT hasGLVersion(int major, int minor, int release = 0);
 // use for both OpenGL and GLX extensions
 bool KWINGLUTILS_EXPORT hasGLExtension(const QByteArray &extension);
 
 // detect OpenGL error (add to various places in code to pinpoint the place)
-bool KWINGLUTILS_EXPORT checkGLError(const char* txt);
+bool KWINGLUTILS_EXPORT checkGLError(const char *txt);
 
 QList<QByteArray> KWINGLUTILS_EXPORT openGLExtensions();
 
@@ -70,14 +58,15 @@ class KWINGLUTILS_EXPORT GLShader
 {
 public:
     enum Flags {
-        NoFlags         = 0,
+        NoFlags = 0,
         ExplicitLinking = (1 << 0)
     };
 
     GLShader(const QString &vertexfile, const QString &fragmentfile, unsigned int flags = NoFlags);
     ~GLShader();
 
-    bool isValid() const  {
+    bool isValid() const
+    {
         return mValid;
     }
 
@@ -86,15 +75,15 @@ public:
 
     bool link();
 
-    int uniformLocation(const char* name);
+    int uniformLocation(const char *name);
 
-    bool setUniform(const char* name, float value);
-    bool setUniform(const char* name, int value);
-    bool setUniform(const char* name, const QVector2D& value);
-    bool setUniform(const char* name, const QVector3D& value);
-    bool setUniform(const char* name, const QVector4D& value);
-    bool setUniform(const char* name, const QMatrix4x4& value);
-    bool setUniform(const char* name, const QColor& color);
+    bool setUniform(const char *name, float value);
+    bool setUniform(const char *name, int value);
+    bool setUniform(const char *name, const QVector2D &value);
+    bool setUniform(const char *name, const QVector3D &value);
+    bool setUniform(const char *name, const QVector4D &value);
+    bool setUniform(const char *name, const QMatrix4x4 &value);
+    bool setUniform(const char *name, const QColor &color);
 
     bool setUniform(int location, float value);
     bool setUniform(int location, int value);
@@ -104,13 +93,13 @@ public:
     bool setUniform(int location, const QMatrix4x4 &value);
     bool setUniform(int location, const QColor &value);
 
-    int attributeLocation(const char* name);
-    bool setAttribute(const char* name, float value);
+    int attributeLocation(const char *name);
+    bool setAttribute(const char *name, float value);
     /**
      * @return The value of the uniform as a matrix
      * @since 4.7
      */
-    QMatrix4x4 getUniformMatrix4x4(const char* name);
+    QMatrix4x4 getUniformMatrix4x4(const char *name);
 
     enum MatrixUniform {
         TextureMatrix = 0,
@@ -138,7 +127,9 @@ public:
     };
 
     enum IntUniform {
-        AlphaToOne,     ///< @deprecated no longer used
+        AlphaToOne, ///< @deprecated no longer used
+        TextureWidth,
+        TextureHeight,
         IntUniformCount
     };
 
@@ -148,16 +139,16 @@ public:
     };
 
     bool setUniform(MatrixUniform uniform, const QMatrix4x4 &matrix);
-    bool setUniform(Vec2Uniform uniform,   const QVector2D &value);
-    bool setUniform(Vec4Uniform uniform,   const QVector4D &value);
-    bool setUniform(FloatUniform uniform,  float value);
-    bool setUniform(IntUniform uniform,    int value);
-    bool setUniform(ColorUniform uniform,  const QVector4D &value);
-    bool setUniform(ColorUniform uniform,  const QColor &value);
+    bool setUniform(Vec2Uniform uniform, const QVector2D &value);
+    bool setUniform(Vec4Uniform uniform, const QVector4D &value);
+    bool setUniform(FloatUniform uniform, float value);
+    bool setUniform(IntUniform uniform, int value);
+    bool setUniform(ColorUniform uniform, const QVector4D &value);
+    bool setUniform(ColorUniform uniform, const QColor &value);
 
 protected:
     GLShader(unsigned int flags = NoFlags);
-    bool loadFromFiles(const QString& vertexfile, const QString& fragmentfile);
+    bool loadFromFiles(const QString &vertexfile, const QString &fragmentfile);
     bool load(const QByteArray &vertexSource, const QByteArray &fragmentSource);
     const QByteArray prepareSource(GLenum shaderType, const QByteArray &sourceCode) const;
     bool compile(GLuint program, GLenum shaderType, const QByteArray &sourceCode) const;
@@ -167,9 +158,9 @@ protected:
 
 private:
     unsigned int mProgram;
-    bool mValid:1;
-    bool mLocationsResolved:1;
-    bool mExplicitLinking:1;
+    bool mValid : 1;
+    bool mLocationsResolved : 1;
+    bool mExplicitLinking : 1;
     int mMatrixLocation[MatrixCount];
     int mVec2Location[Vec2UniformCount];
     int mVec4Location[Vec4UniformCount];
@@ -180,17 +171,15 @@ private:
     friend class ShaderManager;
 };
 
-
 enum class ShaderTrait {
-    MapTexture       = (1 << 0),
-    UniformColor     = (1 << 1),
-    Modulate         = (1 << 2),
+    MapTexture = (1 << 0),
+    UniformColor = (1 << 1),
+    Modulate = (1 << 2),
     AdjustSaturation = (1 << 3),
 };
 
 Q_DECLARE_FLAGS(ShaderTraits, ShaderTrait)
 
-
 /**
  * @short Manager for Shaders.
  *
@@ -219,13 +208,6 @@ public:
      * @return @c true if a shader is bound, @c false otherwise
      */
     bool isShaderBound() const;
-    /**
-     * Is @c true if the environment variable KWIN_GL_DEBUG is set to 1.
-     * In that case shaders are compiled with KWIN_SHADER_DEBUG defined.
-     * @returns @c true if shaders are compiled with debug information
-     * @since 4.8
-     */
-    bool isShaderDebug() const;
 
     /**
      * Pushes the current shader onto the stack and binds a shader
@@ -258,7 +240,7 @@ public:
      * @param fragmentSource The source code of the fragment shader.
      * @return The created shader
      */
-    GLShader *loadShaderFromCode(const QByteArray &vertexSource, const QByteArray &fragmentSource);
+    std::unique_ptr<GLShader> loadShaderFromCode(const QByteArray &vertexSource, const QByteArray &fragmentSource);
 
     /**
      * Creates a custom shader with the given @p traits and custom @p vertexSource and or @p fragmentSource.
@@ -276,13 +258,10 @@ public:
      * @return new generated shader
      * @since 5.6
      */
-    GLShader *generateCustomShader(ShaderTraits traits, const QByteArray &vertexSource = QByteArray(), const QByteArray &fragmentSource = QByteArray());
+    std::unique_ptr<GLShader> generateCustomShader(ShaderTraits traits, const QByteArray &vertexSource = QByteArray(), const QByteArray &fragmentSource = QByteArray());
 
     /**
      * Creates a custom shader with the given @p traits and custom @p vertexFile and or @p fragmentFile.
-     * The file names specified in @p vertexFile and @p fragmentFile are relative paths to the shaders
-     * resource file shipped together with KWin. This means this method can only be used for built-in
-     * effects, for 3rd party effects generateCustomShader should be used.
      *
      * If the @p vertexFile is empty a vertex shader with the given @p traits is generated.
      * If it is not empty the @p vertexFile is used as the source for the vertex shader.
@@ -292,20 +271,16 @@ public:
      * So if both @p vertexFile and @p fragmentFile are provided the @p traits are ignored.
      * If neither are provided a new shader following the @p traits is generated.
      *
+     * If a custom shader stage is provided and core profile is used, the final file path will
+     * be resolved by appending "_core" to the basename.
+     *
      * @param traits The shader traits for generating the shader
      * @param vertexFile optional vertex shader source code to be used instead of shader traits
      * @param fragmentFile optional fragment shader source code to be used instead of shader traits
      * @return new generated shader
      * @see generateCustomShader
-     * @since 5.6
      */
-    GLShader *generateShaderFromResources(ShaderTraits traits, const QString &vertexFile = QString(), const QString &fragmentFile = QString());
-
-    /**
-     * Compiles and tests the dynamically generated shaders.
-     * Returns true if successful and false otherwise.
-     */
-    bool selfTest();
+    std::unique_ptr<GLShader> generateShaderFromFile(ShaderTraits traits, const QString &vertexFile = QString(), const QString &fragmentFile = QString());
 
     /**
      * @return a pointer to the ShaderManager instance
@@ -326,12 +301,10 @@ private:
 
     QByteArray generateVertexSource(ShaderTraits traits) const;
     QByteArray generateFragmentSource(ShaderTraits traits) const;
-    GLShader *generateShader(ShaderTraits traits);
+    std::unique_ptr<GLShader> generateShader(ShaderTraits traits);
 
-    QStack<GLShader*> m_boundShaders;
-    QHash<ShaderTraits, GLShader *> m_shaderHash;
-    bool m_debug;
-    QString m_resourcePath;
+    QStack<GLShader *> m_boundShaders;
+    std::map<ShaderTraits, std::unique_ptr<GLShader>> m_shaderHash;
     static ShaderManager *s_shaderManager;
 };
 
@@ -380,104 +353,99 @@ private:
     GLShader *m_shader;
 };
 
-inline
-ShaderBinder::ShaderBinder(GLShader *shader)
+inline ShaderBinder::ShaderBinder(GLShader *shader)
     : m_shader(shader)
 {
     ShaderManager::instance()->pushShader(shader);
 }
 
-inline
-ShaderBinder::ShaderBinder(ShaderTraits traits)
+inline ShaderBinder::ShaderBinder(ShaderTraits traits)
     : m_shader(nullptr)
 {
     m_shader = ShaderManager::instance()->pushShader(traits);
 }
 
-inline
-ShaderBinder::~ShaderBinder()
+inline ShaderBinder::~ShaderBinder()
 {
     ShaderManager::instance()->popShader();
 }
 
-inline
-GLShader* ShaderBinder::shader()
+inline GLShader *ShaderBinder::shader()
 {
     return m_shader;
 }
 
 /**
- * @short Render target object
+ * @short OpenGL framebuffer object
  *
- * Render target object enables you to render onto a texture. This texture can
- *  later be used to e.g. do post-processing of the scene.
+ * Framebuffer object enables you to render onto a texture. This texture can
+ * later be used to e.g. do post-processing of the scene.
  *
  * @author Rivo Laks <rivolaks@hot.ee>
  */
-class KWINGLUTILS_EXPORT GLRenderTarget
+class KWINGLUTILS_EXPORT GLFramebuffer
 {
 public:
     /**
-     * Constructs a GLRenderTarget
+     * Constructs a GLFramebuffer
      * @since 5.13
      */
-    explicit GLRenderTarget();
+    explicit GLFramebuffer();
 
     /**
-     * Constructs a GLRenderTarget
-     * @param color texture where the scene will be rendered onto
+     * Constructs a GLFramebuffer. Note that ensuring the color attachment outlives
+     * the framebuffer is the responsibility of the caller.
+     *
+     * @param colorAttachment texture where the scene will be rendered onto
      */
-    explicit GLRenderTarget(const GLTexture& color);
-    ~GLRenderTarget();
+    explicit GLFramebuffer(GLTexture *colorAttachment);
 
     /**
-     * Enables this render target.
-     * All OpenGL commands from now on affect this render target until the
-     *  @ref disable method is called
+     * Constructs a wrapper for an already created framebuffer object. The GLFramebuffer
+     * does not take the ownership of the framebuffer object handle.
      */
-    bool enable();
-    /**
-     * Disables this render target, activating whichever target was active
-     *  when @ref enable was called.
-     */
-    bool disable();
+    GLFramebuffer(GLuint handle, const QSize &size);
+    ~GLFramebuffer();
 
     /**
-     * Sets the target texture
-     * @param target texture where the scene will be rendered on
-     * @since 4.8
+     * Returns the framebuffer object handle to this framebuffer object.
      */
-    void attachTexture(const GLTexture& target);
-
+    GLuint handle() const
+    {
+        return mFramebuffer;
+    }
     /**
-     * Detaches the texture that is currently attached to this framebuffer object.
-     * @since 5.13
+     * Returns the size of the color attachment to this framebuffer object.
      */
-    void detachTexture();
-
-    bool valid() const  {
+    QSize size() const
+    {
+        return mSize;
+    }
+    bool valid() const
+    {
         return mValid;
     }
 
-    void setTextureDirty() {
-        mTexture.setDirty();
-    }
-
     static void initStatic();
-    static bool supported()  {
+    static bool supported()
+    {
         return sSupported;
     }
 
     /**
-     * Pushes the render target stack of the input parameter in reverse order.
-     * @param targets The stack of GLRenderTargets
+     * Returns the last bound framebuffer, or @c null if no framebuffer is current.
+     */
+    static GLFramebuffer *currentFramebuffer();
+
+    /**
+     * Pushes the framebuffer stack of the input parameter in reverse order.
+     * @param fbos The stack of GLFramebuffers
      * @since 5.13
      */
-    static void pushRenderTargets(QStack <GLRenderTarget*> targets);
+    static void pushFramebuffers(QStack<GLFramebuffer *> fbos);
 
-    static void pushRenderTarget(GLRenderTarget *target);
-    static GLRenderTarget *popRenderTarget();
-    static bool isRenderTargetBound();
+    static void pushFramebuffer(GLFramebuffer *fbo);
+    static GLFramebuffer *popFramebuffer();
     /**
      * Whether the GL_EXT_framebuffer_blit extension is supported.
      * This functionality is not available in OpenGL ES 2.0.
@@ -488,88 +456,42 @@ public:
     static bool blitSupported();
 
     /**
-     * Blits the content of the current draw framebuffer into the texture attached to this FBO.
+     * Blits from @a source rectangle in the current framebuffer to the @a destination rectangle in
+     * this framebuffer.
+     *
+     * Be aware that framebuffer blitting may not be supported on all hardware. Use blitSupported()
+     * to check whether it is supported.
+     *
+     * The @a source and the @a destination rectangles can have different sizes. The @a filter indicates
+     * what filter will be used in case scaling needs to be performed.
      *
-     * Be aware that framebuffer blitting may not be supported on all hardware. Use blitSupported to check whether
-     * it is supported.
-     * @param source Geometry in screen coordinates which should be blitted, if not specified complete framebuffer is used
-     * @param destination Geometry in attached texture, if not specified complete texture is used as destination
-     * @param filter The filter to use if blitted content needs to be scaled.
      * @see blitSupported
      * @since 4.8
      */
     void blitFromFramebuffer(const QRect &source = QRect(), const QRect &destination = QRect(), GLenum filter = GL_LINEAR);
 
-    /**
-     * Sets the virtual screen size to @p s.
-     * @since 5.2
-     */
-    static void setVirtualScreenSize(const QSize &s) {
-        s_virtualScreenSize = s;
-    }
-
-    /**
-     * Sets the virtual screen geometry to @p g.
-     * This is the geometry of the OpenGL window currently being rendered to
-     * in the virtual geometry space the rendering geometries use.
-     * @see virtualScreenGeometry
-     * @since 5.9
-     */
-    static void setVirtualScreenGeometry(const QRect &g) {
-        s_virtualScreenGeometry = g;
-    }
-
-    /**
-     * The geometry of the OpenGL window currently being rendered to
-     * in the virtual geometry space the rendering system uses.
-     * @see setVirtualScreenGeometry
-     * @since 5.9
-     */
-    static QRect virtualScreenGeometry() {
-        return s_virtualScreenGeometry;
-    }
-
-    /**
-     * The scale of the OpenGL window currently being rendered to
-     *
-     * @returns the ratio between the virtual geometry space the rendering
-     * system uses and the target
-     * @since 5.10
-     */
-    static void setVirtualScreenScale(qreal scale) {
-        s_virtualScreenScale = scale;
-    }
-
-    static qreal virtualScreenScale() {
-        return s_virtualScreenScale;
-    }
-
-
 protected:
-    void initFBO();
-
+    void initFBO(GLTexture *colorAttachment);
 
 private:
+    bool bind();
+
     friend void KWin::cleanupGL();
     static void cleanup();
     static bool sSupported;
     static bool s_blitSupported;
-    static QStack<GLRenderTarget*> s_renderTargets;
-    static QSize s_virtualScreenSize;
-    static QRect s_virtualScreenGeometry;
-    static qreal s_virtualScreenScale;
-    static GLint s_virtualScreenViewport[4];
+    static QStack<GLFramebuffer *> s_fbos;
 
-    GLTexture mTexture;
-    bool mValid;
-
-    GLuint mFramebuffer;
+    GLuint mFramebuffer = 0;
+    QSize mSize;
+    bool mValid = false;
+    bool mForeign = false;
 };
 
 enum VertexAttributeType {
     VA_Position = 0,
     VA_TexCoord = 1,
-    VertexAttributeCount = 2
+    VertexAttributeCount = 2,
 };
 
 /**
@@ -581,10 +503,10 @@ enum VertexAttributeType {
  */
 struct GLVertexAttrib
 {
-    int index;            /** The attribute index */
-    int size;             /** The number of components [1..4] */
-    GLenum type;          /** The type (e.g. GL_FLOAT) */
-    int relativeOffset;   /** The relative offset of the attribute */
+    int index; /** The attribute index */
+    int size; /** The number of components [1..4] */
+    GLenum type; /** The type (e.g. GL_FLOAT) */
+    int relativeOffset; /** The relative offset of the attribute */
 };
 
 /**
@@ -660,7 +582,7 @@ public:
      * @param texcoords The texture coordinates for each vertex.
      * Size must equal 2 * @a numberVertices.
      */
-    void setData(int numberVertices, int dim, const float* vertices, const float* texcoords);
+    void setData(int numberVertices, int dim, const float *vertices, const float *texcoords);
 
     /**
      * Maps an unused range of the data store into the client's address space.
@@ -718,7 +640,7 @@ public:
      * Same as above restricting painting to @a region if @a hardwareClipping is true.
      * It's within the caller's responsibility to enable GL_SCISSOR_TEST.
      */
-    void render(const QRegion& region, GLenum primitiveMode, bool hardwareClipping = false);
+    void render(const QRegion &region, GLenum primitiveMode, bool hardwareClipping = false);
     /**
      * Sets the color the geometry will be rendered with.
      * For legacy rendering glColor is used before rendering the geometry.
@@ -729,7 +651,7 @@ public:
      * @see isUseColor
      * @since 4.7
      */
-    void setColor(const QColor& color, bool enableColor = true);
+    void setColor(const QColor &color, bool enableColor = true);
     /**
      * @return @c true if geometry will be painted with a color, @c false otherwise
      * @see setUseColor
@@ -762,11 +684,11 @@ public:
     void endOfFrame();
 
     /**
-     * Notifies the vertex buffer that we have posted the frame.
+     * Notifies the vertex buffer that we are about to paint a frame.
      *
      * @internal
      */
-    void framePosted();
+    void beginFrame();
 
     /**
      * @internal
@@ -789,31 +711,8 @@ public:
      */
     static GLVertexBuffer *streamingBuffer();
 
-    /**
-     * Sets the virtual screen geometry to @p g.
-     * This is the geometry of the OpenGL window currently being rendered to
-     * in the virtual geometry space the rendering geometries use.
-     * @since 5.9
-     */
-    static void setVirtualScreenGeometry(const QRect &g) {
-        s_virtualScreenGeometry = g;
-    }
-
-    /**
-     * The scale of the OpenGL window currently being rendered to
-     *
-     * @returns the ratio between the virtual geometry space the rendering
-     * system uses and the target
-     * @since 5.11.3
-     */
-    static void setVirtualScreenScale(qreal s) {
-        s_virtualScreenScale = s;
-    }
-
 private:
-    GLVertexBufferPrivate* const d;
-    static QRect s_virtualScreenGeometry;
-    static qreal s_virtualScreenScale;
+    GLVertexBufferPrivate *const d;
 };
 
 } // namespace
